-- Note, most variables and objects use camelCase naming but the panels use upper case CamelCase naming as the objects are sometimes referenced
-- using loadstring() with the name derived from it's string table name which is capitalized. Member values and functions use upper case CamelCase naming.
-- Member functions that extend built in Lua objects (like string.ltrim) are all lowercase. AddCallback and RemoveCallback are left as upper CamelCase to
-- avoid copy/paste errors when reusing components with other plugins
-- I've settled on this naming convention to be most consistent with what Turbine provides as well as the typical lower case camelCase that I tend to use normally.

-- to do
-- remove "setupWindow" as "options" is now a panel in mainWindow. move all remaing setupWindow functionality to mainWindow.

-- turbine client imports
import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

importPath=string.gsub(getfenv(1)._.Name,"%.Main","").."."

-- need localPlayer to test for session play
localPlayer=Turbine.Gameplay.LocalPlayer.GetInstance();
charName=localPlayer:GetName();
-- mmediately test for session play BEFORE importing any other files so that we can safely bail without loading any data and without any handlers
if string.sub(charName,1,1)=="~" then
	-- do not load if the character is a session play character
	Turbine.Shell.WriteLine("Alt Viewer does not support session play characters.");
	return false;
end
selectedCharName=charName;

string.ltrim=function(str)
	return (string.gsub(str, "^%s*(.-)", "%1"))
end
string.rtrim=function(str)
	return (string.gsub(str, "^(.-)%s*$", "%1"))
end
string.trim=function(str)
	return (string.gsub(str, "^%s*(.-)%s*$", "%1"))
end
replaceStr={[128]="A",[129]="A",[130]="A",[131]="A",[132]="A",[133]="A",[134]="AE",[135]="C",[136]="E",[137]="E",[138]="E",[139]="E",[140]="I",[141]="I",[142]="I",[143]="I",[144]="",[145]="N",[146]="O",[147]="O",[148]="O",[149]="O",[150]="O",[151]="",[152]="O",[153]="U",[154]="U",[155]="U",[156]="U",[157]="Y",[158]="Y",[159]="sz",[160]="a",[161]="a",[162]="a",[163]="a",[164]="a",[165]="a",[166]="ae",[167]="c",[168]="e",[169]="e",[170]="e",[171]="e",[172]="i",[173]="i",[174]="i",[175]="i",[176]="o",[177]="n",[178]="o",[179]="o",[180]="o",[181]="o",[182]="o",[183]="",[184]="o",[185]="u",[186]="u",[187]="u",[188]="u",[189]="y",[190]="y",[191]="y"}
string.stripaccent=function(str)
	local ret = "";
	local replace=false
	for i,v in ipairs({str:byte(1,-1)}) do
		if replace then
			replace=false
			if replaceStr[v]~=nil then
				ret=ret..replaceStr[v]
			end
		else
			if (v==195) then
				replace=true
			else
				ret = ret .. string.char(v);
			end
		end
	end
	return ret;
end

import (importPath.."Class");
import (importPath.."VindarPatch");
-- This is the Load/Save patch which accounts for European formats, written by Vindar
-- load resource strings - this MUST occur before any windows that use strings get loaded or there will be no defaults
import (importPath.."ColorPicker");
import (importPath.."Strings");
import (importPath.."Table");
import (importPath.."RadioButtonGroup");
import (importPath.."DropDownList");
import (importPath.."TabControl");

import (importPath.."DebugWindow");
import (importPath.."FontMetric");

language=1;
if (tonumber("1,000")==1) then
	function euroNormalize(value)
		return tonumber((string.gsub(value,"%.",",")));
	end
else
	function euroNormalize(value)
		return tonumber((string.gsub(value,",",".")));
	end
end

-- note, locale is the locale of the running client and is used when comparing client generated strings (such as in chat handlers)
-- where language is based on locale by default but can be overridden by the user to force resource strings to be shown in a different language than the client is set to run in.
-- this allows users running an "EN" client to view the plugin with a "FR/DE" UI or vice-versa.

locale = "en";
if Turbine.Shell.IsCommand("hilfe") then
	locale = "de";
	language=3;
elseif Turbine.Shell.IsCommand("aide") then
	locale = "fr";
	language=2;
end

hudVisible=true; -- assume hud is visible when loading plugins -- I hate this assumption, but Turbine has failed to provide a legitimate means of testing the hud state

displayWidth=Turbine.UI.Display:GetWidth();
displayHeight=Turbine.UI.Display:GetHeight();

charList=PatchDataLoad(Turbine.DataScope.Server, "AltViewerCharList");
if charList==nil then
	charList={};
end
if charList[charName]==nil then
	charList[charName]={};
end
if charList[charName].Gender==nil then
	charList[charName].Gender="M"
end
charList[charName].Race=localPlayer:GetRace();
charList[charName].Class=localPlayer:GetClass();
charList[charName].Level=localPlayer:GetLevel();
charList[charName].Alignment=localPlayer:GetAlignment();
if charList[charName].Alignment==1 then
	charList[charName].Vocation=localPlayer:GetAttributes():GetVocation();
else
	charList[charName].Vocation=nil;
end

settings=PatchDataLoad(Turbine.DataScope.Character, "AltViewerSettings");
if settings==nil then
	settings=PatchDataLoad(Turbine.DataScope.Account, "AltViewerSettings");
end
if settings==nil then
	settings={}
end

if settings.Language==nil then settings.Language=language end
language=settings.Language;
if settings.Debug==nil then settings.Debug=false end;
settings.Debug=false; -- don't allow persistent debugging (to debug initialization this has to be temporarily changed to true)
if settings.ForeColor==nil then
	settings.ForeColor=Turbine.UI.Color(1,.9,.5)
else
	settings.ForeColor=Turbine.UI.Color(euroNormalize(settings.ForeColor.R),euroNormalize(settings.ForeColor.G),euroNormalize(settings.ForeColor.B))
end
if settings.ErrorColor==nil then
	settings.ErrorColor=Turbine.UI.Color(1,0,0)
else
	settings.ErrorColor=Turbine.UI.Color(euroNormalize(settings.ErrorColor.R),euroNormalize(settings.ErrorColor.G),euroNormalize(settings.ErrorColor.B))
end

if settings.TrimColor==nil then
	settings.TrimColor=Turbine.UI.Color(.4,.4,.5)
else
	settings.TrimColor=Turbine.UI.Color(euroNormalize(settings.TrimColor.R),euroNormalize(settings.TrimColor.G),euroNormalize(settings.TrimColor.B))
end
if settings.BackColor==nil then
	settings.BackColor=Turbine.UI.Color(.05,.05,.05)
else
	settings.BackColor=Turbine.UI.Color(euroNormalize(settings.BackColor.R),euroNormalize(settings.BackColor.G),euroNormalize(settings.BackColor.B))
end
if settings.HeadingsColor==nil then
	settings.HeadingsColor=Turbine.UI.Color(1,1,1)
else
	settings.HeadingsColor=Turbine.UI.Color(euroNormalize(settings.HeadingsColor.R),euroNormalize(settings.HeadingsColor.G),euroNormalize(settings.HeadingsColor.B))
end
if settings.ListTextColor==nil then
	settings.ListTextColor=Turbine.UI.Color(1,1,1)
else
	settings.ListTextColor=Turbine.UI.Color(euroNormalize(settings.ListTextColor.R),euroNormalize(settings.ListTextColor.G),euroNormalize(settings.ListTextColor.B))
end

if settings.Font==nil then settings.Font=Turbine.UI.Lotro.Font.Verdana16 end
if settings.Opacity==nil then settings.Opacity=1 end
if settings.LoadMinimized==nil then settings.LoadMinimized=false end
if settings.ShowIcon==nil then settings.ShowIcon=true end
if settings.UseMiniIcon==nil then settings.UseMiniIcon=false end

if settings.DefaultPanel==nil then settings.DefaultPanel="First Available" end

if settings.MainWidth==nil then
	settings.MainWidth=430;
else
	settings.MainWidth=euroNormalize(settings.MainWidth);
	if settings.MainWidth>displayWidth then settings.MainWidth=displayWidth end
end
if settings.MainHeight==nil then
	settings.MainHeight=400;
else
	settings.MainHeight=euroNormalize(settings.MainHeight);
	if settings.MainHeight>displayHeight then settings.MainHeight=displayHeight end
end

if settings.MainLeft==nil then
	settings.MainLeft=.5-(settings.MainWidth/displayWidth)/2;
else
	settings.MainLeft=euroNormalize(settings.MainLeft);
	if settings.MainLeft*displayWidth+settings.MainWidth>displayWidth then settings.MainLeft=1-settings.MainWidth/displayWidth end
end
if settings.MainTop==nil then
	settings.MainTop=.5-(settings.MainHeight/displayHeight)/2;
else
	settings.MainTop=euroNormalize(settings.MainTop);
	if settings.MainTop*displayHeight+settings.MainHeight>displayHeight then settings.MainTop=1-settings.MainHeight/displayHeight end
end

if settings.Panels==nil then
	settings.Panels={}
end

--settings.Panels={}
panelID={}
for k,v in ipairs(settings.Panels) do
	panelID[v.Name]=k
end

if panelID["Recipes"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Recipes",["Show"]=true,["Load"]=false,["StringID"]=5};
	panelID["Recipes"]=order;
end
if panelID["Deeds"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Deeds",["Show"]=true,["Load"]=false,["StringID"]=7};
	panelID["Deeds"]=order;
end
if panelID["Quests"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Quests",["Show"]=true,["Load"]=false,["StringID"]=8};
	panelID["Quests"]=order;
end
if panelID["Rep"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Rep",["Show"]=true,["Load"]=true,["StringID"]=9};
	panelID["Rep"]=order;
end
if panelID["Stats"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Stats",["Show"]=true,["Load"]=true,["StringID"]=10};
	panelID["Stats"]=order;
end
if panelID["Traits"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Traits",["Show"]=true,["Load"]=true,["StringID"]=11};
	panelID["Traits"]=order;
end
if panelID["Notes"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Notes",["Show"]=true,["Load"]=true,["StringID"]=12};
	panelID["Notes"]=order;
end
if panelID["Options"]==nil then
	local order=#settings.Panels+1;
	settings.Panels[order]={["Name"]="Options",["Show"]=true,["Load"]=true,["StringID"]=15};
	panelID["Options"]=order;
end

fontMetricInst=FontMetric();
fontMetricInst:SetFont(settings.Font);

function isPluginLoaded(pluginName)
	local loaded=false;
	local pluginList=Turbine.PluginManager:GetLoadedPlugins();
	local index;
	for k,v in ipairs(pluginList) do
		if v.Name==pluginName then
			loaded=true;
			break;
		end;
	end
	return loaded
end
-- This is the callback mechanism provided by Pengoros, slightly modified to guarantee uniqueness
function AddCallback(Object, Event, Callback)
	if (Object[Event] == nil) then
		Object[Event] = Callback;
	else
		if (type(Object[Event]) == "table") then
			local Exists=false;
			for K,V in ipairs(Object[Event]) do
				if V==Callback then
					Exists=true;
					break;
				end
			end
			if not Exists then
				table.insert(Object[Event], Callback);
			end
		else
			if Object[Event]~=Callback then
				Object[Event] = {Object[Event], Callback};
			end
		end
	end
	return Callback;
end

-- safely remove a Callback without clobbering any extras
function RemoveCallback(Object, Event, Callback)
    if (Object[Event] == Callback) then
        Object[Event] = nil;
    else
        if (type(Object[Event]) == "table") then
            for K,V in ipairs(Object[Event]) do
                if (V == Callback) then
                    table.remove(Object[Event], K);
                    break;
                end
            end
        end
    end
end

import (importPath.."IconWindow");
import (importPath.."MainWindow");

function unloadPlugin()
	SaveData();

	-- have to remove all shell commands before unloading or we get a crash to desktop :(
	if setupWindow~=nil then
		Turbine.Shell.RemoveCommand(setupWindow.ShellCmd)
	end

	-- safely remove all Callbacks
	for k,v in pairs(mainWindow.Panels) do
		if v.RemoveCallbacks ~= nil then
			v.RemoveCallbacks()
		end
	end
end
function SaveData()
	PatchDataSave(Turbine.DataScope.Server,"AltViewerCharList",charList);

	-- save the tooltip settings
	settings.Language=language;
	settings.version=Plugins.AltViewer:GetVersion();
	settings.SWTop=setupWindow:GetTop();
	settings.SWLeft=setupWindow:GetLeft();
	settings.MostRecentPanel=mainWindow.CurrentPanel;

	settings.MainLeft=mainWindow:GetLeft()/displayWidth;
	settings.MainTop=mainWindow:GetTop()/displayHeight;
	settings.MainWidth=mainWindow:GetWidth();
	settings.MainHeight=mainWindow:GetHeight();
	settings.IconTop=iconWindow:GetTop()/displayHeight;
	settings.IconLeft=iconWindow:GetLeft()/displayWidth;
	PatchDataSave(Turbine.DataScope.Character, "AltViewerSettings", settings);
	PatchDataSave(Turbine.DataScope.Account, "AltViewerSettings", settings); -- this allows for easy setup of alts

	for k,v in ipairs(mainWindow.Panels) do
		if v.SaveData~=nil then
			v:SaveData();
		end
	end
end

-- setup/options window
setupWindow = Turbine.UI.Lotro.Window();
setupWindow.Loaded=false;
setupWindow:SetZOrder(1);
local initHeight=410;
local initWidth=445;
if settings.SWTop==nil then settings.SWTop=(displayHeight - initHeight) / 2 end
if settings.SWLeft==nil then settings.SWLeft=(displayWidth - initWidth) / 2 end

if initHeight>displayHeight then initHeight=displayHeight end
if initHeight+settings.SWTop>displayHeight then settings.SWTop=displayHeight-initHeight end
if initWidth>displayWidth then initWidth=displayWidth end
if initWidth+settings.SWLeft>displayWidth then settings.SWLeft=displayWidth-initWidth end

setupWindow:SetPosition( settings.SWLeft, settings.SWTop );
setupWindow:SetHeight(initHeight);
setupWindow:SetWidth(initWidth);
setupWindow:SetOpacity(settings.Opacity);
setupWindow:SetVisible(false);

setupWindow.KeyDown=function(sender, args)
	if ( args.Action == Turbine.UI.Lotro.Action.Escape ) then
		mainWindow:SetVisible(false);
		iconWindow:SetVisible(settings.ShowIcon);
	end
	if ( args.Action == 268435579 ) then -- reposition UI
	end
	if ( args.Action == 268435635 ) then -- toggle HUD
		hudVisible=not hudVisible;
		if hudVisible then
			mainWindow:SetVisible(mainVisible);
			iconWindow:SetVisible(iconVisible and settings.ShowIcon);
		else
			mainVisible=mainWindow:IsVisible();
			mainWindow:SetVisible(false);
			iconVisible=iconWindow:IsVisible();
			iconWindow:SetVisible(false);
		end
	end
end

setupWindow:SetWantsKeyEvents(true);
setupWindow:SetWantsUpdates(true)

setupWindow.ShellCmd=Turbine.ShellCommand();
Turbine.Shell.AddCommand("AltViewer",setupWindow.ShellCmd);

setupWindow.ShellCmd.Execute = function(sender, cmd, args)
	if (string.lower(cmd) == "altviewer") then
		if (string.lower(args)=="show") then
			iconWindow:SetVisible(false);
			mainWindow:SetVisible(true);
		elseif (string.lower(args)=="hide") then
			mainWindow:SetVisible(false);
			if settings.ShowIcon then
				iconWindow:SetVisible(true);
			end
		elseif (string.lower(args)=="toggle") then
			if mainWindow:IsVisible() then
				mainWindow:SetVisible(false);
				if settings.ShowIcon then
					iconWindow:SetVisible(true);
				end
			else
				iconWindow:SetVisible(false);
				mainWindow:SetVisible(true);
			end
		elseif (string.lower(args)=="debug on") then
			settings.Debug=true;
			if debugWindow==nil then
				debugWindow=DebugWindow();
			end
			debugWindow:SetVisible(true);
			Turbine.Shell.WriteLine("Debug is now: "..tostring(settings.Debug))
		elseif (string.lower(args)=="debug off") then
			settings.Debug=false;
			if debugWindow~=nil then
				debugWindow:SetVisible(false);
			end
			Turbine.Shell.WriteLine("Debug is now: "..tostring(settings.Debug))
		elseif (string.lower(args)=="debug toggle") then
			settings.Debug=not settings.Debug;
			if debugWindow~=nil then
				debugWindow:SetVisible(settings.Debug);
			end
			Turbine.Shell.WriteLine("Debug is now: "..tostring(settings.Debug))
		elseif (string.lower(args)=="debug") then
			Turbine.Shell.WriteLine("Debug is: "..tostring(settings.Debug))
		else
			Turbine.Shell.WriteLine("usage:");
			Turbine.Shell.WriteLine("  /AltViewer show");
			Turbine.Shell.WriteLine("       Shows the main window.");
			Turbine.Shell.WriteLine("  /AltViewer hide");
			Turbine.Shell.WriteLine("       Hides the main window.");
			Turbine.Shell.WriteLine("  /AltViewer toggle");
			Turbine.Shell.WriteLine("       Toggles the main window.");
		end
	end
end
setupWindow.ShellCmd.GetHelp = function(sender, cmd)
	return("usage:\n  /AltViewer show\n       Shows the main window.\n  /AltViewer hide\n       Hides the main window.\n  /AltViewer toggle\n       Toggles the main window.");
end

setupWindow.ShellCmd.GetShortHelp = function(sender, cmd)
	return("AltViewer - usage: /AltViewer show | hide | toggle");
end

setupWindow.Update=function( sender, args )
	if (Plugins.AltViewer ~= nil) and (not setupWindow.Loaded) then
		setupWindow:SetWantsUpdates(false)
		initCharList();
		setupWindow.Loaded=true;
		Turbine.Shell.WriteLine(Resource[language][2].." "..Plugins.AltViewer:GetVersion().." "..Resource[language][3]);
		Plugins.AltViewer.Unload = function(setupWindow,sender,args)
			unloadPlugin();
		end
		if isPluginLoaded("AltViewerReloader") then
			Turbine.PluginManager.UnloadScriptState("AltViewerReloader");
			-- now restore the setup window state
			for k,v in ipairs(settings.Panels) do
				if v.Name=="Options" then
					local sender=mainWindow.IconStrip:GetItem(k);
					local args={};
					sender.MouseClick(sender,args);
					break;
				end
			end
			OptionsPanel.PanelTab.Button:MouseClick();
		end
	end
end

function initCharList()

	-- initialize the character list
--	mainWindow.CharList:AddItem(Resource[language][12]);
--	mainWindow.CharList:AddItem(Resource[language][9]);
	local selIndex=1;
	local tmpList={};
	for k,v in pairs(charList) do
		table.insert(tmpList,k);
	end
	table.sort(tmpList)
	for k,v in ipairs(tmpList) do
		if v==charName then selIndex=k end
		mainWindow.CharList:AddItem(v);
	end
	mainWindow.CharList:SetSelectedIndex(selIndex);
	mainWindow.CharList:SelectedIndexChanged();
end
